from __future__ import annotations

import select
import socket
from types import TracebackType
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing_extensions import Self

from multicast_expert import LOCALHOST_IPV4, LOCALHOST_IPV6, os_multicast
from multicast_expert.utils import (
    IPv4Or6Address,
    MulticastExpertError,
    get_interface_ips,
    is_windows,
    validate_mcast_ip,
)


class McastRxSocket:
    """
    Class to wrap a socket that receives from one or more multicast groups.
    """

    def __init__(
        self,
        addr_family: int,
        mcast_ips: list[str],
        port: int,
        iface_ip: str | None = None,
        iface_ips: list[str] | None = None,
        source_ips: list[str] | None = None,
        timeout: float | None = None,
        blocking: bool | None = None,
        enable_external_loopback: bool = False,
    ):
        """
        Create a socket which receives UDP datagrams over multicast.

        The socket must be opened (e.g. using a with statement) before it can be used.

        Note: This socket can only receive multicast traffic, not regular unicast traffic.

        :param addr_family: Sets IPv4 or IPv6 operation.  Either socket.AF_INET or socket.AF_INET6.
        :param mcast_ips: List of all possible multicast IPs that this socket can receive from.
        :param port: The port to listen on.
        :param iface_ips: Interface IPs that this socket receives from.  If left as None, multicast_expert will
            attempt to listen on all (non-loopback) interfaces discovered on your machine.  Be careful, this default
            may not be desired in many cases.  See the docs for details.
        :param iface_ip: Legacy alias for iface_ips.  If this is given and iface_ips is not, this adds the
            given interface IP to iface_ips.
        :param source_ips: Optional, list of source addresses to restrict the multicast subscription to.  The
            OS will drop messages not from one of these IPs, and may even use special IGMPv3 source-specific
            subscription packets to ask for only those specific multicasts from switches/routers.
            This option is only supported for IPv4, currently no major OS supports it with IPv6.
        :param timeout: Timeout for receive operations using this socket.  See settimeout() for how
            this value is processed.
        :param blocking: Legacy alias for timeout.  If set to True, timeout is set to None (block forever).  If set to
            False, timeout is set to 0 (nonblocking).
        :param enable_external_loopback: Enable loopback of multicast packets sent on external interfaces. If and only
            if this option is set to True, McastRxSockets will be able to receive packets sent by McastTxSockets open on
            the same address, port, and interface.
        """
        self.addr_family = addr_family
        self.mcast_ips = mcast_ips
        self.port = port
        self.source_ips = source_ips

        self.is_opened = False

        # blocking overrides timeout if set
        if blocking is not None:
            self.timeout: float | None = None if blocking else 0.0
        else:
            self.timeout = timeout

        # Handle legacy iface_ip argument if given
        self.iface_ips: list[str]
        if iface_ip is not None:
            if iface_ips is not None:
                message = "Both iface_ips and iface_ip may not be specified at the same time!"
                raise MulticastExpertError(message)

            self.iface_ips = [iface_ip]

        elif iface_ips is None:
            self.iface_ips = get_interface_ips(addr_family == socket.AF_INET, addr_family == socket.AF_INET6)

            # Don't include the localhost IPs when listening on all interfaces, as that would cause
            # us to receive all mcasts sent by the current machine.
            if self.addr_family == socket.AF_INET6 and LOCALHOST_IPV6 in self.iface_ips:
                self.iface_ips.remove(LOCALHOST_IPV6)
            if self.addr_family == socket.AF_INET and LOCALHOST_IPV4 in self.iface_ips:
                self.iface_ips.remove(LOCALHOST_IPV4)

            if len(self.iface_ips) == 0:
                message = "Unable to discover any listenable interfaces on this machine."
                raise MulticastExpertError(message)
        else:
            self.iface_ips = iface_ips

        # Resolve the interfaces now.  This prevents having to do this relatively expensive call
        # multiple times later.
        self.iface_infos = {}
        for ip in self.iface_ips:
            try:
                self.iface_infos[ip] = os_multicast.get_iface_info(ip)
            except KeyError as ex:
                message = f"Interface IP {ip} does not seem to correspond to a valid interface.  Valid interfaces: {', '.join(get_interface_ips())}"
                raise MulticastExpertError(message) from ex

        # Sanity check multicast addresses
        for mcast_ip in self.mcast_ips:
            validate_mcast_ip(mcast_ip, self.addr_family)

        # Sanity check source_ips
        self.is_source_specific = not (source_ips is None or len(source_ips) == 0)
        if self.is_source_specific and self.addr_family == socket.AF_INET6:
            message = "Source-specific multicast currently cannot be used with IPv6!"
            raise MulticastExpertError(message)

        self.enable_external_loopback = enable_external_loopback

    def __enter__(self) -> Self:
        if self.is_opened:
            message = "Attempt to open an McastRxSocket that is already open!"
            raise MulticastExpertError(message)

        # Create the sockets and set options
        self.sockets = []

        # On Windows, we have to create a socket and bind it for each interface address, then subscribe
        # to all multicast addresses on each of those sockets
        # On Unix, we need one socket bound to each multicast address.
        if is_windows:
            for iface_ip in self.iface_ips:
                new_socket = socket.socket(family=self.addr_family, type=socket.SOCK_DGRAM)
                new_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

                new_socket.bind((iface_ip, self.port))

                if self.is_source_specific:
                    os_multicast.add_source_specific_memberships(
                        new_socket, self.mcast_ips, cast(list[str], self.source_ips), self.iface_infos[iface_ip]
                    )
                else:
                    os_multicast.add_memberships(
                        new_socket, self.mcast_ips, self.iface_infos[iface_ip], self.addr_family
                    )

                # On Windows, by default, sent packets are looped back to local sockets on the same interface, even for interfaces
                # that are not loopback.Change this by disabling IP_MULTICAST_LOOP unless the loopback interface is used or
                # if enable_external_loopback is set.
                # Note: multicast_expert submitted a PR to clarify this in the Windows docs, and it was accepted!
                loop_enabled = self.enable_external_loopback or iface_ip in (LOCALHOST_IPV4, LOCALHOST_IPV6)
                if self.addr_family == socket.AF_INET:
                    new_socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, loop_enabled)
                else:
                    new_socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_LOOP, loop_enabled)

                self.sockets.append(new_socket)
        else:
            for mcast_ip in self.mcast_ips:
                # For IPv6 on Unix, we need to create one socket for each mcast_ip - iface_ip permutation.
                # For IPv4, on the systems I tested at least, you can get away with subscribing to multiple
                # interfaces on one socket.
                if self.addr_family == socket.AF_INET6:
                    iface_ip_groups = [[iface_ip] for iface_ip in self.iface_ips]
                else:
                    iface_ip_groups = [self.iface_ips]

                for iface_ips_this_group in iface_ip_groups:
                    new_socket = socket.socket(family=self.addr_family, type=socket.SOCK_DGRAM)
                    new_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

                    if self.addr_family == socket.AF_INET6:
                        # Note: for Unix IPv6, need to specify the scope ID in the bind address in order for link-local mcast addresses to work.
                        new_socket.bind((mcast_ip, self.port, 0, self.iface_infos[iface_ips_this_group[0]].iface_idx))
                    else:
                        new_socket.bind((mcast_ip, self.port))

                    for iface_ip in iface_ips_this_group:
                        if self.is_source_specific:
                            os_multicast.add_source_specific_memberships(
                                new_socket, [mcast_ip], cast(list[str], self.source_ips), self.iface_infos[iface_ip]
                            )
                        else:
                            os_multicast.add_memberships(
                                new_socket, [mcast_ip], self.iface_infos[iface_ip], self.addr_family
                            )

                    self.sockets.append(new_socket)

        self.is_opened = True

        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc: BaseException | None, traceback: TracebackType | None
    ) -> None:
        if not self.is_opened:
            message = "Attempt to open an McastRxSocket that is already open!"
            raise MulticastExpertError(message)

        # Close socket
        for sock in self.sockets:
            sock.close()
        self.is_opened = False

    def recvfrom(self, bufsize: int = 4096, flags: int = 0) -> tuple[bytes, IPv4Or6Address] | None:
        """
        Receive a UDP packet from the socket, returning the bytes and the sender address.

        This respects the current blocking and timeout settings.

        The "bufsize" and "flags" arguments have the same meaning as the arguments to socket.recv(), see the
        manual for that function for details.

        :param bufsize: Maximum amount of data to be received at once.
        :param flags: Flags that will be passed to the OS.

        :return: Tuple of (bytes, address).  For IPv4, address is a tuple of IP address (str) and port number.
            For IPv6, address is a tuple of IP address (str), port number, flow info (int), and scope ID (int).
            If no packets were received (nonblocking mode or timeout), None is returned.
        """
        # Use select() to find a socket that is ready for reading
        read_list, write_list, exception_list = select.select(self.sockets, [], [], self.timeout)

        if len(read_list) == 0:
            # No data to read
            return None

        # Since we only want to return one packet at a time, just pick the first readable socket.
        return cast(tuple[bytes, IPv4Or6Address], read_list[0].recvfrom(bufsize, flags))

    def recv(self, bufsize: int = 4096, flags: int = 0) -> bytes | None:
        """
        Receive a UDP packet from the socket, returning the bytes.

        This respects the current blocking and timeout settings.

        Note: If you need information about the sender of the packet, use recvfrom() instead.

        The "bufsize" and "flags" arguments have the same meaning as the arguments to socket.recv(), see the
        manual for that function for details.

        :param bufsize: Maximum amount of data to be received at once.
        :param flags: Flags that will be passed to the OS.

        :return: Bytes received.
        """
        packet_and_addr = self.recvfrom(bufsize, flags)
        if packet_and_addr is None:
            return None
        else:
            return packet_and_addr[0]

    def filenos(self) -> list[int]:
        """
        Get a list of the socket file descriptor(s) used by this socket.

        You can use this with the select module to implement blocking I/O on multiple different multicast sockets.

        :return: socket file descriptor(s) used by this socket.
        """
        return [socket.fileno() for socket in self.sockets]

    def settimeout(self, timeout: float | None) -> None:
        """
        Set the timeout on socket operations.

        :param timeout: The timeout. Possible values:
            - Number > 0: Receiving packets will abort if more than timeout seconds elapse while waiting for a packet.
            - 0: Socket will be put in nonblocking mode
            - None: Socket will block forever (the default)

        """
        self.timeout = timeout



from __future__ import annotations

import socket
from types import TracebackType
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self

from multicast_expert import LOCALHOST_IPV4, LOCALHOST_IPV6, os_multicast
from multicast_expert.utils import (
    IPv4Or6Address,
    MulticastExpertError,
    get_default_gateway_iface_ip,
    get_interface_ips,
    is_mac,
    is_windows,
    validate_mcast_ip,
)


class McastTxSocket:
    """
    Class to wrap a socket that sends to one or more multicast groups.
    """

    def __init__(
        self,
        addr_family: int,
        mcast_ips: list[str],
        iface_ip: str | None = None,
        ttl: int = 1,
        enable_external_loopback: bool = False,
    ):
        """
        Create a socket which transmits UDP datagrams over multicast.

        The socket must be opened (e.g. using a with statement) before it can be used.

        :param addr_family: Sets IPv4 or IPv6 operation.  Either socket.AF_INET or socket.AF_INET6.
        :param mcast_ips: List of all possible multicast IPs that this socket can send to.
        :param iface_ip: Interface IP that this socket sends on.  If left as None, multicast_expert will
            attempt to guess an interface by using the interface your default gateway is on (aka
            the one your PC uses to access the internet).  Be careful, this default may not be desired
            in many cases.  See the docs for details.
        :param ttl: Time-to-live parameter to set on the packets sent by this socket, AKA hop limit for ipv6.
            This controls the number of routers that the packets may pass through until they are dropped.  The
            default value of 1 prevents the multicast packets from passing through any routers.
        :param enable_external_loopback: Enable receiving multicast packets sent over external interfaces. If and only
            if this option is set to True, McastRxSockets will be able to receive packets sent by McastTxSockets open on
            the same address, port, and interface.
        """
        self.addr_family = addr_family
        self.iface_ip = iface_ip
        self.mcast_ips = mcast_ips
        self.mcast_ips_set = set(mcast_ips)  # Used for checking IPs in send()
        self.ttl = ttl
        self.is_opened = False

        if self.iface_ip is None:
            self.iface_ip = get_default_gateway_iface_ip(self.addr_family)

            if self.iface_ip is None:
                message = "iface_ip not specified but unable to determine the default gateway on this machine"
                raise MulticastExpertError(message)

        # Resolve the interface
        try:
            self.iface_info = os_multicast.get_iface_info(self.iface_ip)
        except KeyError as ex:
            message = f"iface_ip {self.iface_ip} does not seem to correspond to a valid interface.  Valid interfaces: {', '.join(get_interface_ips())}"
            raise MulticastExpertError(message) from ex

        # Sanity check multicast addresses
        for mcast_ip in self.mcast_ips:
            validate_mcast_ip(mcast_ip, self.addr_family)

        self.enable_external_loopback = enable_external_loopback

    def __enter__(self) -> Self:
        if self.is_opened:
            message = "Attempt to open an McastTxSocket that is already open!"
            raise MulticastExpertError(message)

        # Open the socket and set options
        self.socket = socket.socket(family=self.addr_family, type=socket.SOCK_DGRAM)

        # Note: for Unix IPv6, need to specify the scope ID in the bind address
        if self.addr_family == socket.AF_INET6 and not is_windows:
            self.socket.bind((self.iface_ip, 0, 0, self.iface_info.iface_idx))
        else:
            self.socket.bind((self.iface_ip, 0))  # Bind to any available port

        # Use the IP_MULTICAST_IF option to set the interface to use.
        os_multicast.set_multicast_if(self.socket, self.mcast_ips, self.iface_info, self.addr_family)

        # Now set the time-to-live (thank goodness, this is the same on all platforms)
        if self.addr_family == socket.AF_INET:
            self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, self.ttl)
        else:  # IPv6
            self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_HOPS, self.ttl)

        # On Unix, we need to disable multicast loop if we do not want sent packets to get looped back to local
        # sockets on the same interface.
        if not is_windows:
            # Enable loopback if enable_external_loopback is set or if using a loopback interface on Mac.
            # Otherwise, disable loopback.
            enable_loopback = self.enable_external_loopback or (
                is_mac and (self.iface_ip in (LOCALHOST_IPV4, LOCALHOST_IPV6))
            )

            if self.addr_family == socket.AF_INET:
                self.socket.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_LOOP, enable_loopback)
            else:
                self.socket.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_LOOP, enable_loopback)

        self.is_opened = True

        return self

    def __exit__(
        self, exc_type: type[BaseException] | None, exc: BaseException | None, traceback: TracebackType | None
    ) -> None:
        if not self.is_opened:
            message = "Attempt to close an McastTxSocket that is already closed!"
            raise MulticastExpertError(message)

        # Close socket
        self.socket.close()
        self.is_opened = False

    def sendto(self, tx_bytes: bytes, address: IPv4Or6Address) -> None:
        """
        Send a UDP datagram containing the given bytes out of the socket to the given destination address.

        :param tx_bytes: Bytes to send
        :param address: Tuple of the destination multicast address and the destination port
        """
        if address[0] not in self.mcast_ips_set:
            message = f"The given destination address ({address[0]}) was not one of the addresses given for this McastTxSocket to transmit to!"
            raise MulticastExpertError(message)

        self.socket.sendto(tx_bytes, address)

    def fileno(self) -> int:
        """
        Get the file descriptor for this socket.

        :return: File descriptor for this socket.
        """
        return self.socket.fileno()

    def getsockname(self) -> tuple[str, int]:
        """
        Get the local IP and port that this socket bound itself to.

        :return: Tuple of local IP and local port
        """
        return self.socket.getsockname()  # type: ignore[no-any-return]



import platform
import socket
import sys

# Constants section
# ------------------------------------------------------------------------------------------------------

# Add missing source-specific multicast constants
# (cpython bug #89415, fixed in python 3.12)
if sys.version_info < (3, 12, 0):
    if platform.system() == "Windows":
        if not hasattr(socket, "IP_ADD_SOURCE_MEMBERSHIP"):
            socket.IP_ADD_SOURCE_MEMBERSHIP = 15  # type: ignore[attr-defined]
        if not hasattr(socket, "IP_DROP_SOURCE_MEMBERSHIP"):
            socket.IP_DROP_SOURCE_MEMBERSHIP = 16  # type: ignore[attr-defined]
    elif platform.system() == "Darwin":
        if not hasattr(socket, "IP_ADD_SOURCE_MEMBERSHIP"):
            socket.IP_ADD_SOURCE_MEMBERSHIP = 70  # type: ignore[attr-defined]
        if not hasattr(socket, "IP_DROP_SOURCE_MEMBERSHIP"):
            socket.IP_DROP_SOURCE_MEMBERSHIP = 71  # type: ignore[attr-defined]
    else:  # Assume Linux/Unix
        if not hasattr(socket, "IP_ADD_SOURCE_MEMBERSHIP"):
            socket.IP_ADD_SOURCE_MEMBERSHIP = 39  # type: ignore[attr-defined]
        if not hasattr(socket, "IP_DROP_SOURCE_MEMBERSHIP"):
            socket.IP_DROP_SOURCE_MEMBERSHIP = 40  # type: ignore[attr-defined]

# Localhost / loopback addresses, for convenience
LOCALHOST_IPV4 = "127.0.0.1"
LOCALHOST_IPV6 = "::1"


from multicast_expert.rx_socket import McastRxSocket as McastRxSocket
from multicast_expert.tx_socket import McastTxSocket as McastTxSocket
from multicast_expert.utils import IPv4Or6Address as IPv4Or6Address
from multicast_expert.utils import MulticastExpertError as MulticastExpertError
from multicast_expert.utils import get_default_gateway_iface_ip as get_default_gateway_iface_ip
from multicast_expert.utils import get_default_gateway_iface_ip_v4 as get_default_gateway_iface_ip_v4
from multicast_expert.utils import get_default_gateway_iface_ip_v6 as get_default_gateway_iface_ip_v6
from multicast_expert.utils import get_interface_ips as get_interface_ips
from multicast_expert.utils import validate_mcast_ip as validate_mcast_ip



from __future__ import annotations

import ipaddress
import platform
import socket
from typing import Union

import netifaces

# docsig: disable

is_windows = platform.system() == "Windows"
is_mac = platform.system() == "Darwin"


# Type which can represent an IPv4 or an IPv6 address.
# For IPv4, address is a tuple of IP address (str) and port number.
# For IPv6, address is a tuple of IP address (str), port number, flow info (int), and scope ID (int).
# NOTE: Since we still support python 3.9, we cannot use new style type annotations here. We have to use the
# old Union annotation which works in 3.9
IPv4Or6Address = Union[tuple[str, int], tuple[str, int, int, int]]


# Exception class for this library
class MulticastExpertError(RuntimeError):
    pass


def get_interface_ips(include_ipv4: bool = True, include_ipv6: bool = True) -> list[str]:
    """
    Use this function to get a list of all the interface IP addresses available on this machine.

    Should be useful for generating help menus / info messages / etc.
    This is a thin wrapper around the netifaces library's functionality.

    :param include_ipv4: If true, IPv4 addresses will be included in the results
    :param include_ipv6: If true, IPv6 addresses will be included in the results

    :return: List of the interface IP of every interface on your machine, as a string.
    """
    ip_list: list[str] = []
    for interface in netifaces.interfaces():
        all_addresses: list[dict[str, str]] = []
        addresses_at_each_level = netifaces.ifaddresses(interface)

        if include_ipv4:
            # Note: Check needed because some interfaces do not have an ipv4 or ipv6 address
            if netifaces.AF_INET in addresses_at_each_level:
                all_addresses.extend(addresses_at_each_level[netifaces.AF_INET])

        if include_ipv6:
            if netifaces.AF_INET6 in addresses_at_each_level:
                all_addresses.extend(addresses_at_each_level[netifaces.AF_INET6])

        ip_list.extend(address_dict["addr"] for address_dict in all_addresses)

    return ip_list


def get_default_gateway_iface_ip_v6() -> str | None:
    """
    Get the IP address of the interface that connects to the default IPv6 gateway.

    If this cannot be determined, None is returned.
    """
    return get_default_gateway_iface_ip(netifaces.AF_INET6)


def get_default_gateway_iface_ip_v4() -> str | None:
    """
    Get the IP address of the interface that connects to the default IPv4 gateway.

    If this cannot be determined, None is returned.
    """
    return get_default_gateway_iface_ip(netifaces.AF_INET)


def get_default_gateway_iface_ip(addr_family: int) -> str | None:
    """
    Get the IP address of the interface that connects to the default gateway of the given addr_family.

    If this cannot be determined, None is returned.
    """
    # Enumerate all gateways using netifaces
    try:
        gateways = netifaces.gateways()
    except OSError:
        return None

    # If it can, it will identify one of those as the default gateway for traffic.
    # If not, return none.
    if not "default" in gateways:
        return None
    if not addr_family in gateways["default"]:
        return None

    default_gateway = gateways["default"][addr_family]
    default_gateway_iface = default_gateway[1]  # element 1 is the iface name, per the docs

    # Now, use the interface name to get the IP address of that interface
    interface_addresses: dict[int, list[dict[str, str]]] = netifaces.ifaddresses(default_gateway_iface)
    if addr_family not in interface_addresses:
        return None
    return interface_addresses[addr_family][0]["addr"]


def validate_mcast_ip(mcast_ip: str, addr_family: int) -> None:
    """
    Validate that the given mcast_ip is a valid multicast address in the given addr family (IPv4 or IPv6).

    An exception is thrown if validation fails.
    """
    address_obj = ipaddress.ip_address(mcast_ip)
    if not address_obj.is_multicast:
        message = f"mcast_ip {mcast_ip} is not a multicast address!"
        raise MulticastExpertError(message)

    if address_obj is ipaddress.IPv4Address and addr_family == socket.AF_INET6:
        message = f"mcast_ip {mcast_ip} is IPv4 but this is an AF_INET6 socket!"
        raise MulticastExpertError(message)

    if address_obj is ipaddress.IPv6Address and addr_family == socket.AF_INET:
        message = f"mcast_ip {mcast_ip} is IPv6 but this is an AF_INET socket!"
        raise MulticastExpertError(message)
